//**********************************************************
//Copyright 2011 Fethallah Benmansour
//
//Licensed under the Apache License, Version 2.0 (the "License"); 
//you may not use this file except in compliance with the License. 
//You may obtain a copy of the License at
//
//http://www.apache.org/licenses/LICENSE-2.0 
//
//Unless required by applicable law or agreed to in writing, software 
//distributed under the License is distributed on an "AS IS" BASIS, 
//WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
//See the License for the specific language governing permissions and 
//limitations under the License.
//**********************************************************


#ifndef __itkStructureTensorRecursiveGaussianImageFilter_txx
#define __itkStructureTensorRecursiveGaussianImageFilter_txx

#include "itkStructureTensorRecursiveGaussianImageFilter.h"
#include "itkImageRegionIteratorWithIndex.h"

namespace itk
{

/**
 * Constructor
 */
template <typename TInputImage, typename TOutputImage >
StructureTensorRecursiveGaussianImageFilter<TInputImage,TOutputImage>
::StructureTensorRecursiveGaussianImageFilter()
{
  m_NormalizeAcrossScale = false;

  unsigned int imageDimensionMinus1 = static_cast<int>(ImageDimension)-1;
  if( ImageDimension > 1)
    {
    m_SmoothingFilters.resize(imageDimensionMinus1);
		m_TensorComponentSmoothingFilters.resize(ImageDimension);
    }


  for( unsigned int i = 0; i<imageDimensionMinus1; i++ )
    {
    m_SmoothingFilters[ i ] = GaussianFilterType::New();
    m_SmoothingFilters[ i ]->SetOrder( GaussianFilterType::ZeroOrder );
    m_SmoothingFilters[ i ]->SetNormalizeAcrossScale( m_NormalizeAcrossScale );
    m_SmoothingFilters[ i ]->ReleaseDataFlagOn();
    }
	
	for( unsigned int i = 0; i < ImageDimension; i++ )
    {
    m_TensorComponentSmoothingFilters[i] = GaussianFilterType::New();
		m_TensorComponentSmoothingFilters[i]->SetOrder( GaussianFilterType::ZeroOrder );
		m_TensorComponentSmoothingFilters[i]->SetNormalizeAcrossScale( m_NormalizeAcrossScale );
		m_TensorComponentSmoothingFilters[i]->ReleaseDataFlagOn();
		m_TensorComponentSmoothingFilters[i]->SetDirection(i);
		if(i > 0)
			{
			m_TensorComponentSmoothingFilters[i]->SetInput( m_TensorComponentSmoothingFilters[i-1]->GetOutput() );
			}
		}
    
  m_DerivativeFilter = DerivativeFilterType::New();
  m_DerivativeFilter->SetOrder( DerivativeFilterType::FirstOrder );
  m_DerivativeFilter->SetNormalizeAcrossScale( m_NormalizeAcrossScale );

  m_DerivativeFilter->SetInput( this->GetInput() );

  if( ImageDimension > 1 )
    {
    m_SmoothingFilters[0]->SetInput( m_DerivativeFilter->GetOutput() );
    }

  for( unsigned int i = 1; i<imageDimensionMinus1; i++ )
    {
    m_SmoothingFilters[ i ]->SetInput(
      m_SmoothingFilters[i-1]->GetOutput() );
    }

  m_ImageAdaptor = OutputImageAdaptorType::New();

  m_Sigma = 1.0;
  m_Rho   = 1.0;
  this->SetSigma( 1.0 );
  this->SetRho( 1.0 );

}

/**
 * Set value of Sigma
 */
template <typename TInputImage, typename TOutputImage>
void
StructureTensorRecursiveGaussianImageFilter<TInputImage,TOutputImage>
::SetSigma( RealType sigma )
{
  m_Sigma = sigma;
  for( unsigned int i = 0; i<ImageDimension-1; i++ )
    {
    m_SmoothingFilters[ i ]->SetSigma( sigma );
    }
  m_DerivativeFilter->SetSigma( sigma );

  this->Modified();

}


/**
 * Set value of Rho
 */
template <typename TInputImage, typename TOutputImage>
void
StructureTensorRecursiveGaussianImageFilter<TInputImage,TOutputImage>
::SetRho( RealType rho )
{
		m_Rho = rho;
		for( unsigned int i = 0; i < ImageDimension; i++ )
			{
			m_TensorComponentSmoothingFilters[i]->SetSigma( rho );
			}
		this->Modified();
}


/**
 * Set Normalize Across Scale Space
 */
template <typename TInputImage, typename TOutputImage>
void
StructureTensorRecursiveGaussianImageFilter<TInputImage,TOutputImage>
::SetNormalizeAcrossScale( bool normalize )
{

  m_NormalizeAcrossScale = normalize;

  for( unsigned int i = 0; i<ImageDimension-1; i++ )
    {
    m_SmoothingFilters[ i ]->SetNormalizeAcrossScale( normalize );
    }
  m_DerivativeFilter->SetNormalizeAcrossScale( normalize );

  this->Modified();

}

template <typename TInputImage, typename TOutputImage>
void
StructureTensorRecursiveGaussianImageFilter<TInputImage,TOutputImage>
::GenerateInputRequestedRegion() throw(InvalidRequestedRegionError)
{
  // call the superclass' implementation of this method. this should
  // copy the output requested region to the input requested region
  Superclass::GenerateInputRequestedRegion();

  // This filter needs all of the input
  typename StructureTensorRecursiveGaussianImageFilter<TInputImage,TOutputImage>
           ::InputImagePointer image = const_cast<InputImageType *>( this->GetInput() );
  image->SetRequestedRegion( this->GetInput()->GetLargestPossibleRegion() );
}

template <typename TInputImage, typename TOutputImage>
void
StructureTensorRecursiveGaussianImageFilter<TInputImage,TOutputImage>
::EnlargeOutputRequestedRegion(DataObject *output)
{
  TOutputImage *out = dynamic_cast<TOutputImage*>(output);

  if (out)
    {
    out->SetRequestedRegion( out->GetLargestPossibleRegion() );
    }
}

/**
 * Compute filter for Gaussian kernel
 */
template <typename TInputImage, typename TOutputImage >
void
StructureTensorRecursiveGaussianImageFilter<TInputImage,TOutputImage >
::GenerateData(void)
{
  // Create a process accumulator for tracking the progress of this
  // minipipeline
  ProgressAccumulator::Pointer progress = ProgressAccumulator::New();
  progress->SetMiniPipelineFilter(this);

  unsigned int imageDimensionMinus1 = static_cast<int>(ImageDimension)-1;
  // Compute the contribution of each filter to the total progress.
  const double weight = 1.0 / ( ImageDimension * ImageDimension );

  for( unsigned int i = 0; i<ImageDimension-1; i++ )
    {
    progress->RegisterInternalFilter( m_SmoothingFilters[i], weight );
    }

  progress->RegisterInternalFilter( m_DerivativeFilter, weight );
  progress->ResetProgress();

  const typename TInputImage::ConstPointer   inputImage( this->GetInput() );

  m_ImageAdaptor->SetImage( this->GetOutput() );

  m_ImageAdaptor->SetLargestPossibleRegion(
    inputImage->GetLargestPossibleRegion() );

  m_ImageAdaptor->SetBufferedRegion(
    inputImage->GetBufferedRegion() );

  m_ImageAdaptor->SetRequestedRegion(
    inputImage->GetRequestedRegion() );

  m_ImageAdaptor->Allocate();

  m_DerivativeFilter->SetInput( inputImage );

  for( unsigned int dim=0; dim < ImageDimension; dim++ )
    {
    unsigned int i=0;
    unsigned int j=0;
    while(  i< imageDimensionMinus1)
      {
      if( i == dim )
        {
        j++;
        }
      m_SmoothingFilters[ i ]->SetDirection( j );
      i++;
      j++;
      }
    m_DerivativeFilter->SetDirection( dim );

    GaussianFilterPointer lastFilter;

    if( ImageDimension > 1 ) 
      {
      int imageDimensionMinus2 = static_cast<int>(ImageDimension)-2;
      lastFilter = m_SmoothingFilters[imageDimensionMinus2];
      lastFilter->Update();
      }
    else
      {
      m_DerivativeFilter->Update();
      }


    progress->ResetFilterProgressAndKeepAccumulatedProgress();

    // Copy the results to the corresponding component
    // on the output image of vectors
    m_ImageAdaptor->SelectNthElement( dim );

    typename RealImageType::Pointer derivativeImage =
      lastFilter->GetOutput();

    ImageRegionIteratorWithIndex< RealImageType > it(
      derivativeImage,
      derivativeImage->GetRequestedRegion() );

    ImageRegionIteratorWithIndex< OutputImageAdaptorType > ot(
      m_ImageAdaptor,
      m_ImageAdaptor->GetRequestedRegion() );

    const RealType spacing = inputImage->GetSpacing()[ dim ];

    it.GoToBegin();
    ot.GoToBegin();
    while( !it.IsAtEnd() )
      {
      ot.Set( it.Get() / spacing );
      ++it;
      ++ot;
      }
    }

  // Calculate the outer (diadic) product of the gradient.
  ImageRegionIteratorWithIndex< OutputImageType > ottensor(
    this->GetOutput(),
    this->GetOutput()->GetRequestedRegion() );

  ImageRegionIteratorWithIndex< OutputImageType > itgradient(
    this->GetOutput(),
    this->GetOutput()->GetRequestedRegion() );

  const unsigned int numberTensorElements = (ImageDimension*(ImageDimension+1))/2;

  std::vector<InternalRealType> tmp( numberTensorElements );

  ottensor.GoToBegin();
  itgradient.GoToBegin();
  while( !itgradient.IsAtEnd() )
    {
    unsigned int count = 0;
    for ( unsigned int j = 0; j < ImageDimension; ++j)
      {
      for (int k = j; k < ImageDimension; ++k)
        {
        tmp[count++] = itgradient.Get()[j]*itgradient.Get()[k];
        }
      }
    
    for (unsigned int j = 0; j < numberTensorElements; j++)
      {
      ottensor.Value()[j] = tmp[j];
      }
      
    ++itgradient;
    ++ottensor;
    }
    
    
  typedef typename itk::Image<InternalRealType, ImageDimension> ComponentImageType;
    
    
  for(unsigned int i =0; i < numberTensorElements; i++)
    {
    typename ComponentImageType::Pointer componentImage = ComponentImageType::New();
    componentImage->SetLargestPossibleRegion( inputImage->GetLargestPossibleRegion() );
    componentImage->SetBufferedRegion( inputImage->GetBufferedRegion() );
    componentImage->SetRequestedRegion( inputImage->GetRequestedRegion() );
    componentImage->Allocate();
    
    ImageRegionIteratorWithIndex< ComponentImageType > compit(
                                                              componentImage,
                                                              componentImage->GetRequestedRegion());
    
    ottensor.GoToBegin();
    compit.GoToBegin();
    while( !compit.IsAtEnd() )
      {
      compit.Value() = ottensor.Get()[i];
      ++compit;
      ++ottensor;
      }
    
    m_TensorComponentSmoothingFilters[0]->SetInput(componentImage);
    m_TensorComponentSmoothingFilters[ImageDimension-1]->Update();
    
    ImageRegionIteratorWithIndex< ComponentImageType > smoothedCompIt(
                                                                      m_TensorComponentSmoothingFilters[ImageDimension-1]->GetOutput(),
                                                                      m_TensorComponentSmoothingFilters[ImageDimension-1]->GetOutput()->GetRequestedRegion());
    
    ottensor.GoToBegin();
    smoothedCompIt.GoToBegin();
    while( !ottensor.IsAtEnd() )
      {
      ottensor.Value()[i] = smoothedCompIt.Get();  
      ++smoothedCompIt;
      ++ottensor;
      }
    
    }
		
		/** cleaning the filters so that if the input image changes ... TO COMMENT MORE */
		for(unsigned int i = 0; i < ImageDimension-1; i++)
			{
			m_SmoothingFilters[i]->GetOutput()->SetRequestedRegion
						( typename GaussianFilterType::OutputImageType::RegionType() );
			}
		for(unsigned int i = 0; i < ImageDimension; i++)
			{	
			m_TensorComponentSmoothingFilters[i]->GetOutput()->SetRequestedRegion
						( typename GaussianFilterType::OutputImageType::RegionType() );
			}
		m_DerivativeFilter->GetOutput()->SetRequestedRegion
						( typename GaussianFilterType::OutputImageType::RegionType() );				
		
}


template <typename TInputImage, typename TOutputImage>
void
StructureTensorRecursiveGaussianImageFilter<TInputImage,TOutputImage>
::PrintSelf(std::ostream& os, Indent indent) const
{
  Superclass::PrintSelf(os,indent);
  os << "NormalizeAcrossScale: " << m_NormalizeAcrossScale << std::endl;
}


} // end namespace itk

#endif
